#!/usr/bin/python

import datetime

from . import misc

class RateLimiter(misc.Object):
	def handle_request(self, request, handler, minutes, limit):
		return RateLimiterRequest(self.backend, request, handler,
			minutes=minutes, limit=limit)


class RateLimiterRequest(misc.Object):
	prefix = "ratelimit"

	def init(self, request, handler, minutes, limit):
		self.request = request
		self.handler = handler

		# Save the limits
		self.minutes = minutes
		self.limit   = limit

		self.now = datetime.datetime.utcnow()

		# Fetch the current counter value from the cache
		self.counter = self.get_counter()

		# Increment the rate-limiting counter
		self.increment_counter()

		# Write the header if we are not limited
		if not self.is_ratelimited():
			self.write_headers()

	def is_ratelimited(self):
		"""
			Returns True if the request is prohibited by the rate limiter
		"""
		# The client is rate-limited when more requests have been
		# received than allowed.
		return self.counter >= self.limit

	def get_counter(self):
		"""
			Returns the number of requests that have been done in
			recent time.
		"""
		keys = self.get_keys_to_check()

		res = self.memcache.get_multi(keys)
		if res:
			return sum((int(e) for e in res.values()))

		return 0

	def write_headers(self):
		# Send the limit to the user
		self.handler.set_header("X-Rate-Limit-Limit", self.limit)

		# Send the user how many requests are left for this time window
		self.handler.set_header("X-Rate-Limit-Remaining",
			self.limit - self.counter)

		expires = self.now + datetime.timedelta(seconds=self.expires_after)
		self.handler.set_header("X-Rate-Limit-Reset", expires.strftime("%s"))

	def get_key(self):
		key_prefix = self.get_key_prefix()

		return "%s-%s" % (key_prefix, self.now.strftime("%Y-%m-%d-%H:%M"))

	def get_keys_to_check(self):
		key_prefix = self.get_key_prefix()

		keys = []
		for minute in range(self.minutes + 1):
			when = self.now - datetime.timedelta(minutes=minute)

			key = "%s-%s" % (key_prefix, when.strftime("%Y-%m-%d-%H:%M"))
			keys.append(key)

		return keys

	def get_key_prefix(self):
		return "-".join((self.prefix, self.request.host, self.request.path,
			self.request.method, self.request.remote_ip,))

	def increment_counter(self):
		key = self.get_key()

		# Add the key or increment if it already exists
		if not self.memcache.add(key, "1", self.expires_after):
			self.memcache.incr(key)

	@property
	def expires_after(self):
		"""
			Returns the number of seconds after which the counter has reset.
		"""
		return (self.minutes + 1) * 60
